"An example with a broken composer."

#%%
from pathlib import Path
from random import choice, random

import pandas as pd
import numpy as np
import plotly.express as px
import math

from fn_graph import Composer

prices = [random() * 100_000 + 50000 for _ in range(10)]

f = (
    Composer()
    .update(
        wave=lambda frequency, amplitude: pd.DataFrame(dict(x=range(501))).assign(
            y=lambda df: amplitude
            * np.cos(df.x / 500 * math.pi * 3)
            * np.sin(df.x / 500 * math.pi * frequency)
        ),
        plot=lambda wave: px.line(wave, x="x", y="y"),
        broken=lambda wave, missing: wave,
    )
    .update_parameters(frequency=(float, 1), amplitude=(float, 1))
)

# %%



import sys
from pathlib import Path
import random
from random import choice, random
import pandas as pd

sys.path.insert(0, str(Path(__file__).parent.parent.parent.resolve()))

from fn_graph import Composer

# Introductory example


def a():
    return 5


def b(a):
    return a * 5


def c(a, b):
    return a * b


composer = Composer().update(a, b, c)

# Call any result
composer.c()  # 125
composer.a()  # 5

composer.graphviz().render("intro.gv", format="png")
# Some pure functions


def get_car_prices():
    df = pd.DataFrame(
        dict(
            model=[choice(["corolla", "beetle", "ferrari"]) for _ in range(10)],
            price=[random() * 100_000 + 50000 for _ in range(10)],
        )
    )
    return df


def get_mean_car_price(car_prices, season="summer"):
    if season != "summer":
        return car_prices.price.mean() / 2
    else:
        return car_prices.price.mean()


def get_cheaper_cars(car_prices, your_car_price):
    df = car_prices
    return df[df.price < your_car_price]


def get_savings_on_cheaper_cars(cheaper_cars, mean_car_price):
    return cheaper_cars.assign(savings=lambda df: mean_car_price - df.price)


# Compose the functions
g = (
    Composer()
    .update_without_prefix(
        "get_",
        get_car_prices,
        get_cheaper_cars,
        get_mean_car_price,
        get_savings_on_cheaper_cars,
    )
    .update_parameters(your_car_price=100_000)
)
# Calculate a function
print(g.calculate(["savings_on_cheaper_cars"]))

# Or many functions
print(g.calculate(["savings_on_cheaper_cars", "mean_car_price"]))

# Or use the shorthand

print(g.savings_on_cheaper_cars())

# Override a default argument
g.update(season=lambda: "spring").cheaper_cars()
print(g.savings_on_cheaper_cars())

# Lets create some more functions and create a partial composition


def burger_savings(savings_df, price_of_a_burger):
    return savings_df.assign(burgers_saved=lambda df: df.savings / price_of_a_burger)


f = Composer().update(price_of_a_burger=lambda: 100).update(burger_savings)

# We can see we have some errors
print(list(f.check()))


# Lets compose g and f together
# We also add another function to bridge the name

h = g.update_from(f).update(
    savings_df=lambda savings_on_cheaper_cars: savings_on_cheaper_cars
)

# see no errors
print(list(h.check()))


# Lets calculate the result while logging the progress
def log_progress(event, details):
    time = f'{details.get("time"): .5f}' if "time" in details else ""
    print(f"{event:20s} {details.get('fname','')[:50]:50} {time}")


print(
    h.update(season=lambda: "winter").calculate(
        ["burger_savings"], progress_callback=log_progress
    )["burger_savings"]
)


# You can't have cycles
cycles = Composer().update(a=lambda c: c, b=lambda a: a, c=lambda b: b)
print(list(cycles.check()))


# Lets create a new composer using namespaces
# We also link the ideas that don't have matching names
# We use "__" to separate namespace components
h = (
    Composer()
    .update_namespaces(cars=g, burgers=f)
    .link(burgers__savings_df="cars__savings_on_cheaper_cars")
)


# There is a lot you can do with name spaces.
h.calculate(["burgers__burger_savings"])

# If you draw the graph you can see the namespaces
h.graphviz()

g.calculate(["savings_on_cheaper_cars"], intermediates=False)

# %%



import time


def now():
    return time.perf_counter()


class Profiler:
    def __init__(self):
        self.start = {}
        self.end = {}

    def __call__(self, event_type, details):
        name = details.get("name")
        instruction = {
            "start_calculation": ("start", "calculation", "preparation"),
            "prepared_calculation": ("end", "calculation", "preparation"),
            "start_step": ("start", "step", name),
            "end_step": ("end", "step", name),
            "start_function": ("start", "execution", name),
            "end_function": ("end", "execution", name),
            "start_cache_retrieval": ("start", "cache_retrieval", name),
            "end_cache_retrieval": ("end", "cache_retrieval", name),
            "start_cache_store": ("start", "cache_store", name),
            "end_cache_store": ("end", "cache_store", name),
        }.get(event_type)

        if instruction:
            event, category, name = instruction
            store = getattr(self, event)
            store[(category, name)] = now()

    def results(self):
        def period(key):
            return self.end.get(key, 0) - self.start.get(key, 0)

        def function_profile(name):
            total = period(("step", name))
            execution = period(("execution", name))
            retrieval = period(("cache_retrieval", name))
            store = period(("cache_store", name))

            overhead = total - execution - retrieval - store
            return dict(
                total=total,
                overhead=overhead,
                execution=execution,
                cache_retrieval=retrieval,
                cache_store=store,
            )

        names = [name for event, name in self.start.keys() if event == "step"]

        return dict(
            startup=dict(
                preparation=dict(
                    preparation=period(("calculation", "preparation")),
                    total=period(("calculation", "preparation")),
                )
            ),
            functions={name: function_profile(name) for name in names},
        )



"""
A simple example showing basic functionality.
"""
#%%
from random import choice, random

import pandas as pd
import plotly.express as px
from fn_graph import Composer

prices = [random() * 100_000 + 50000 for _ in range(10)]


def get_car_prices():
    df = pd.DataFrame(
        dict(
            model=[choice(["corolla", "beetle", "ferrari"]) for _ in range(10)],
            price=prices,
        )
    )

    return df


def get_mean_car_price(car_prices, season="summer"):
    if season != "summer":
        return car_prices.price.mean() / 2
    else:
        return car_prices.price.mean()


def get_cheaper_cars(car_prices, your_car_price):
    df = car_prices
    return df[df.price < your_car_price]


def get_savings_on_cheaper_cars(cheaper_cars, mean_car_price):
    return cheaper_cars.assign(savings=lambda df: mean_car_price - df.price)


def get_burger_savings(savings_on_cheaper_cars, price_of_a_burger):
    return savings_on_cheaper_cars.assign(
        burgers_saved=lambda df: df.savings / price_of_a_burger
    )


def get_savings_histogram(burger_savings):
    return px.histogram(burger_savings, x="burgers_saved")


f = (
    Composer()
    .update_without_prefix(
        "get_",
        get_car_prices,
        get_cheaper_cars,
        get_mean_car_price,
        get_savings_on_cheaper_cars,
        get_burger_savings,
        get_savings_histogram,
    )
    .update_parameters(your_car_price=(int, 100_000), price_of_a_burger=(float, 100))
)



"A basic example of namespaces."
#%%
from fn_graph import Composer

# Namespaces example


def data():
    return 5


def b(data, factor):
    return data * factor


def c(b):
    return b


def combined_result(child_one__c, child_two__c):
    pass


child = Composer().update(b, c)
parent = (
    Composer()
    .update_namespaces(child_one=child, child_two=child)
    .update(data, combined_result)
    .update_parameters(child_one__factor=3, child_two__factor=5)
)

# %%
parent.graphviz()

# %%

# Link example


def calculated_factor(data):
    return data / 2


factor_calc = Composer()
factoring = Composer().update(calculated_factor)

linked_parent = (
    Composer()
    .update_namespaces(child_one=child, factoring=factoring)
    .update(data)
    .link(child_one__factor="factoring__calculated_factor")
)

f = linked_parent

# %%



from fn_graph import Composer


def test_simple_parameters():
    composer = (
        Composer().update(c=lambda a, b: a + b).update_parameters(a=1, b=(int, 2))
    )
    assert composer.c() == 3


def test_default_arguments():
    composer = Composer().update(c=lambda a, b=3: a + b).update_parameters(a=1)
    assert composer.c() == 4
    composer = Composer().update(c=lambda a, b=3: a + b).update_parameters(a=1, b=2)
    assert composer.c() == 3


def test_var_args():
    a = 1
    b = 2
    result = a + sum(a * 2 for i in range(10)) + b + sum(a * 5 for i in range(5))

    def d(a, *d_, b, **c_):
        return a + sum(d_) + b + sum(c_.values())

    composer = (
        Composer()
        .update_parameters(a=1, b=2)
        .update(**{f"c_{i}": lambda a: a * 2 for i in range(10)})
        .update(**{f"d_{i}": lambda a: a * 5 for i in range(5)})
        .update(d)
    )

    assert composer.d() == result


def test_empty_var_args():
    a = 1
    b = 2
    result = a + sum(a * 2 for i in range(10)) + b

    def d(a, *d_, b, **c_):
        return a + sum(d_) + b + sum(c_.values())

    composer = (
        Composer()
        .update_parameters(a=1, b=2)
        .update(**{f"c_{i}": lambda a: a * 2 for i in range(10)})
        .update(d)
    )

    assert composer.d() == result


def test_empty_kwargs():
    a = 1
    b = 2
    result = a + b + sum(a * 5 for i in range(5))

    def d(a, *d_, b, **c_):
        return a + sum(d_) + b + sum(c_.values())

    composer = (
        Composer()
        .update_parameters(a=1, b=2)
        .update(**{f"d_{i}": lambda a: a * 5 for i in range(5)})
        .update(d)
    )

    assert composer.d() == result



from .utils import compare_composer_results, generate_random_graph
from random import choice
from fn_graph.tests.large_graph import large_graph
from fn_graph import Composer
from ..calculation import get_execution_instructions

root = large_graph
composers = [root, root.cache(), root.development_cache(__name__)]


def test_static_cache_equality():
    root = large_graph
    composers = [root, root.cache(), root.development_cache(__name__)]
    compare_composer_results(root, composers)


def test_cache_clear():
    for i in range(5):
        for composer in composers:
            composer.call(choice(list(composer.dag().nodes())))
            composer.cache_clear()


def test_cache_graphviz():
    for composer in composers:
        composer.call(choice(list(composer.dag().nodes())))
        composer.cache_invalidate(choice(list(composer.dag().nodes())))
        composer.cache_graphviz()


def test_same_graphs():
    for i in range(10):
        root = generate_random_graph()
        composers = [root, root, root]
        compare_composer_results(root, composers)


def test_random_graphs():
    for i in range(10):
        root = generate_random_graph()
        composers = [root, root.cache()]
        compare_composer_results(root, composers)


def test_cache_invalidation():
    composer = Composer().update_parameters(a=1).cache()
    assert composer.a() == 1
    composer = composer.update_parameters(a=2)
    assert composer.a() == 2



"A simple machine learing example."

from fn_graph import Composer
import sklearn, sklearn.datasets, sklearn.svm, sklearn.linear_model, sklearn.metrics
from sklearn.model_selection import train_test_split
import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pylab as plt


def iris():
    """
    Load the classic iris dataset
    """
    return sklearn.datasets.load_iris()


def data(iris):
    """
    Pull out the data as pandas DataFrame
    """
    df_train = pd.DataFrame(
        iris.data, columns=["feature{}".format(i) for i in range(4)]
    )
    return df_train.assign(y=iris.target)


def investigate_data(data):
    """
    Check for any visual correlations using seaborn
    """
    return sns.pairplot(data, hue="y")


def preprocess_data(data, do_preprocess):
    """
    Preprocess the data by scaling depending on the parameter

    We make sure we don't mutate the data because that is better practice.
    """
    processed = data.copy()
    if do_preprocess:
        processed.iloc[:, :-1] = sklearn.preprocessing.scale(processed.iloc[:, :-1])
    return processed


def split_data(preprocess_data):
    """
    Split the data into test and train sets
    """
    return dict(
        zip(
            ("training_features", "test_features", "training_target", "test_target"),
            train_test_split(preprocess_data.iloc[:, :-1], preprocess_data["y"]),
        )
    )


# This is done verbosely purpose, but it could be more concise


def training_features(split_data):
    return split_data["training_features"]


def training_target(split_data):
    return split_data["training_target"]


def test_features(split_data):
    return split_data["test_features"]


def test_target(split_data):
    return split_data["test_target"]


def model(training_features, training_target, model_type):
    """
    Train the model
    """
    if model_type == "ols":
        model = sklearn.linear_model.LogisticRegression()
    elif model_type == "svm":
        model = sklearn.svm.SVC()
    else:
        raise ValueError("invalid model selection, choose either 'ols' or 'svm'")
    model.fit(training_features, training_target)
    return model


def predictions(model, test_features):
    """
    Make some predictions foo the test data
    """
    return model.predict(test_features)


def classification_metrics(predictions, test_target):
    """
    Show some standard classification metrics
    """
    return sklearn.metrics.classification_report(test_target, predictions)


def plot_confusion_matrix(
    cm, target_names, title="Confusion matrix", cmap=plt.cm.Blues
):
    """
    Plots a confusion matrix using matplotlib. 
    
    This is just a regular function that is not in the composer. 
    Shamelessly taken from https://scikit-learn.org/0.15/auto_examples/model_selection/plot_confusion_matrix.html
    """

    plt.imshow(cm, interpolation="nearest", cmap=cmap)
    plt.title(title)
    plt.colorbar()
    tick_marks = np.arange(len(target_names))
    plt.xticks(tick_marks, target_names, rotation=45)
    plt.yticks(tick_marks, target_names)
    plt.tight_layout()
    plt.ylabel("True label")
    plt.xlabel("Predicted label")
    return plt.gcf()


def confusion_matrix(predictions, test_target):
    """
    Show the confusion matrix
    """
    cm = sklearn.metrics.confusion_matrix(test_target, predictions)
    return plot_confusion_matrix(cm, ["setosa", "versicolor", "virginica"])


f = (
    Composer()
    .update_parameters(
        # Parameter controlling the model type (ols, svc)
        model_type="olm",
        # Parameter enabling data preprocessing
        do_preprocess=True,
    )
    .update(
        iris,
        data,
        preprocess_data,
        investigate_data,
        split_data,
        training_features,
        training_target,
        test_features,
        test_target,
        model,
        predictions,
        classification_metrics,
        confusion_matrix,
    )
)



from fn_graph import Composer


def test_basic_links():
    composer = Composer().update(a=lambda: 5, c=lambda b: b * 2).link(b="a")

    assert composer.c() == 10


def test_updated_from_links():
    composer_a = Composer().update(a=lambda: 5, c=lambda b: b * 2).link(b="a")
    composer_b = Composer().update_from(composer_a).update(d=lambda b: b * 3)
    assert composer_b.d() == 15


def test_updated_from_namespaces():
    composer_child = Composer().update(a=lambda: 5, c=lambda b: b * 2).link(b="a")
    composer = (
        Composer()
        .update_namespaces(x=composer_child, y=composer_child)
        .link(outer_x="x__c", outer_y="y__c")
        .update(final=lambda outer_x, outer_y: outer_x + outer_y)
    )

    assert composer.final() == 20


def test_call_link():
    composer = Composer().update(a=lambda: 5, c=lambda b: b * 2).link(b="a")

    assert composer.b() == 5



from __future__ import annotations

from collections import defaultdict, namedtuple
from functools import reduce
from inspect import Parameter, signature
import inspect
from itertools import groupby
from logging import getLogger
from typing import Any, Callable, List

import graphviz
import networkx as nx
from littleutils import strip_required_prefix, strip_required_suffix

from fn_graph.calculation import NodeInstruction, get_execution_instructions
from .caches import DevelopmentCache, SimpleCache, NullCache
from .calculation import calculate

log = getLogger(__name__)


ComposerTestResult = namedtuple("TestResult", "name passed exception")
"""
The results of Composer.run_tests()
"""

# TODO cleverer use of caching - stop unnecessary reads


class Composer:
    """
    A function composer is responsible for orchestrating the composition
    and execution of graph of functions.

    Pure free functions are added to the composer, the names of the function
    arguments are used to determine how those functions are wired up into
    a directed acyclic graph.
    """

    def __init__(
        self,
        *,
        _functions=None,
        _parameters=None,
        _cache=None,
        _tests=None,
        _source_map=None,
    ):
        # These are namespaced
        self._functions = _functions or {}
        self._tests = _tests or {}

        self._cache = _cache or NullCache()
        self._parameters = _parameters or {}
        self._source_map = _source_map or {}

    def _copy(self, **kwargs):
        return type(self)(
            **{
                **dict(
                    _functions=self._functions,
                    _cache=self._cache,
                    _parameters=self._parameters,
                    _tests=self._tests,
                    _source_map=self._source_map,
                ),
                **kwargs,
            }
        )

    def update(self, *args: Callable, **kwargs: Callable) -> Composer:
        """
        Add functions to the composer.

        Args:
            args: Positional arguments use the __name__ of the function as the reference
        in the graph.
            kwargs: Keyword arguments use the key as the name of the function in the graph.

        Returns:
            A new composer with the functions added.
        """
        args_with_names = {arg.__name__: arg for arg in args}
        all_args = {**self._functions, **args_with_names, **kwargs}

        for argname, argument in all_args.items():
            if not callable(argument):
                raise Exception(
                    f"Argument '{argname}' is not a function or callable. All arguments must be callable."
                )

        return self._copy(_functions=all_args)

    def update_without_prefix(
        self, prefix: str, *functions: Callable, **kwargs: Callable
    ) -> Composer:
        """
        Given a prefix and a list of (named) functions, this adds the functions
        to the composer but first strips the prefix from their name. This is very
        useful to stop name shadowing.

        Args:
            prefix: The prefix to strip off the function names
            functions: functions to add while stripping the prefix
            kwargs: named functions to add
        Returns:
            A new composer with the functions added
        """

        args_with_names = {
            strip_required_prefix(arg.__name__, prefix): arg for arg in functions
        }
        return self.update(**args_with_names, **kwargs)

    def update_without_suffix(self, suffix: str, *functions, **kwargs) -> Composer:
        """
        Given a suffix and a list of (named) functions, this adds the functions
        to the composer but first strips the suffix from their name. This is very
        useful to stop name shadowing.

        Args:
            suffix: The suffix to strip off the function names
            functions: functions to add while stripping the suffix
            kwargs: named functions to add
        Returns:
            A new composer with the functions added
        """

        args_with_names = {
            strip_required_suffix(arg.__name__, suffix): arg for arg in functions
        }
        return self.update(**args_with_names, **kwargs)

    def update_from(self, *composers: Composer) -> Composer:
        """
        Create a new composer with all the functions from this composer
        as well as the the passed composers.

        Args:
            composers: The composers to take functions from

        Returns:
            A new Composer with all the input composers functions added.
        """
        return reduce(
            lambda x, y: x.update(**y._functions)
            .update_parameters(**y._parameters)
            .update_tests(**y._tests),
            [self, *composers],
        )

    def update_namespaces(self, **namespaces: Composer) -> Composer:
        """
        Given a group of keyword named composers, create a series of functions
        namespaced by the keywords and drawn from the composers' functions.

        Args:
            namespaces: Composers that will be added at the namespace that corresponds \
            to the arguments key

        Returns:
            A new Composer with all the input composers functions added as namespaces.
        """
        return self._copy(
            **{
                arg: {
                    **getattr(self, arg),
                    **{
                        "__".join([namespace, k]): value
                        for namespace, composer in namespaces.items()
                        for k, value in getattr(composer, arg).items()
                    },
                }
                for arg in ["_functions", "_parameters"]
            }
        )

    def update_parameters(self, **parameters: Any) -> Composer:
        """
        Allows you to pass static parameters to the graph, they will be exposed as callables.
        """

        hydrated_parameters = {}
        for key, parameter in parameters.items():
            if isinstance(parameter, tuple):
                hydrated_parameters[key] = parameter
            else:
                type_ = self._parameters.get(key, (type(parameter), None))[0]
                hydrated_parameters[key] = (type_, parameter)

        def serve_parameter(key, type_, value):
            def parameter():
                cast_value = value
                if isinstance(cast_value, int) and issubclass(type_, float):
                    cast_value = float(value)

                if not isinstance(cast_value, type_):
                    raise Exception(f"Parameter '{key}' is not of type {type_}")
                return cast_value

            return parameter

        # Have to capture the value eagerly
        return self._copy(
            _parameters={**self._parameters, **hydrated_parameters}
        ).update(
            **{
                key: serve_parameter(key, type_, value)
                for key, (type_, value) in hydrated_parameters.items()
            }
        )

    def update_tests(self, **tests) -> Composer:
        """
        Adds tests to the composer.

        A test is a function that should check a property of the calculations
        results and raise an exception if they are not met.

        The work exactly the same way as functions in terms of resolution of
        arguments. They are run with the run_test method.
        """
        return self._copy(_tests={**self._tests, **tests})

    def link(self, **kwargs):
        """
        Create a symlink between an argument name and a function output.
        This is a convenience method. For example:

        `f.link(my_unknown_argument="my_real_function")`

        is the same as

        `f.update(my_unknown_argument= lambda my_real_function: my_real_function)`

        """

        def make_link_fn(source):
            fn = eval(f"lambda {source}: {source}")
            fn._is_fn_graph_link = True
            return fn

        fns = {key: make_link_fn(source) for key, source in kwargs.items()}
        return self.update(**fns)

    def functions(self):
        """
        Dictionary of the functions
        """
        return self._functions

    def parameters(self):
        """
        Dictionary of the parameters of the form {key: (type, value)}
        """
        return self._parameters

    def check(self, outputs=None):
        """
        Returns a generator of errors if there are any errors in the function graph.
        """
        if outputs is None:
            dag = self.dag()
        else:
            dag = self.ancestor_dag(outputs)

        cycles = list(nx.simple_cycles(dag))
        if cycles:
            yield dict(
                type="cycle",
                message=f"Cycle found [{', '.join(cycles[0])}]. The function graph must be acyclic.",
            )

        for unbound_fn, calling_fns in self.subgraph(dag.nodes())._unbound().items():
            yield dict(
                type="unbound",
                message=f"Unbound function '{unbound_fn}' required.",
                function=unbound_fn,
                referers=calling_fns,
            )

    def calculate(
        self, outputs, perform_checks=True, intermediates=False, progress_callback=None
    ):
        return calculate(
            self, outputs, perform_checks, intermediates, progress_callback
        )

    def run_tests(self):
        """
        Run all the composer tests.

        Returns:
            A generator of ComposerTestResults(name, passed, exception).
        """

        all_referenced_functions = [
            self._resolve_predecessor(tname, pname)
            for tname, fn in self._tests.items()
            for pname in signature(fn).parameters
        ]

        results = self.calculate(all_referenced_functions)

        for tname, fn in self._tests.items():

            arguments = {
                pname: results[self._resolve_predecessor(tname, pname)]
                for pname in signature(fn).parameters
            }

            try:
                fn(**arguments)
                test_result = ComposerTestResult(
                    name=tname, passed=True, exception=None
                )
            except Exception as e:
                test_result = ComposerTestResult(name=tname, passed=False, exception=e)

            yield test_result

    def call(self, output):
        """
        A convenience method to calculate a single output
        """
        return self.calculate([output])[output]

    def precalculate(self, outputs):
        """
        Create a new Composer where the results of the given functions have
        been pre-calculated.

        :param outputs: list of the names of the functions to pre-calculate
        :rtype: A composer
        """
        results = self.calculate(outputs)
        return self.update(
            **{k: (lambda x: (lambda: x))(v) for k, v in results.items()}
        )

    def __getattr__(self, name):
        """
        Allow composed functions to be easily called.
        """
        name = name.replace(".", "__")
        if name in self._functions:
            return lambda: self.calculate([name])[name]
        else:
            raise AttributeError(
                f"{self.__class__.__name__} object has no attribute '{name}', nor any composed function '{name}'."
            )

    def raw_function(self, name):
        """
        Access a raw function in the composer by name. Returns None if not found.
        """
        return self._functions.get(name)

    def dag(self):
        """
        Generates the DAG representing the function graph.

        :rtype: a networkx.DiGraph instance with function names as nodes
        """

        G = nx.DiGraph()

        for key in self._functions:
            G.add_node(key)

        for key, fn in self._functions.items():
            predecessors = self._resolve_predecessors(key)
            G.add_edges_from([(resolved, key) for _, resolved in predecessors])

        return G

    def ancestor_dag(self, outputs):
        """
        A dag of all the ancestors of the given outputs, i.e. the functions that must be calculated
        to for the given outputs.
        """
        full_dag = self.dag()
        ancestors = set(outputs) | {
            pred for output in outputs for pred in nx.ancestors(full_dag, output)
        }
        return full_dag.subgraph(ancestors)

    def subgraph(self, function_names):
        """
        Given a collection of function names this will create a new
        composer that only consists of those nodes.
        """
        return self._copy(
            _functions={
                k: self._functions[k] for k in function_names if k in self._functions
            },
            _parameters={
                k: v for k, v in self._parameters.items() if k in function_names
            },
        )

    def cache(self, backend=None) -> Composer:
        """
        Create a new composer with a given cache backend.

        By default this is a SimpleCache.
        """
        backend = backend or SimpleCache()
        return self._copy(_cache=backend)

    def development_cache(self, name, cache_dir=None) -> Composer:
        """
        Create a new composer with a development cache setup
        """
        return self.cache(DevelopmentCache(name, cache_dir))

    def cache_clear(self):
        """
        Clear the cache
        """
        for key in self.dag():
            self._cache.invalidate(self, key)

    def cache_invalidate(self, *nodes: List[str]):
        """
        Invalidate the cache for all nodes affected  by the
        given nodes (the descendants).
        """
        to_invalidate = set()
        for node in nodes:
            to_invalidate.update(nx.descendants(self.dag(), node))
            to_invalidate.add(node)

        for key in to_invalidate:
            self._cache.invalidate(self, key)

    def cache_graphviz(self, outputs=(), **kwargs):
        """
        Display a graphviz with the cache invalidated nodes highlighted.
        """
        instructions = get_execution_instructions(self, self.dag(), outputs)

        filter = self.ancestor_dag(outputs).nodes() if outputs else None

        def get_node_styles(instruction):
            return {
                NodeInstruction.IGNORE: dict(fillcolor="green"),
                NodeInstruction.RETRIEVE: dict(fillcolor="orange"),
                NodeInstruction.CALCULATE: dict(fillcolor="red"),
            }[instruction]

        # TODO: Be a bit more careful about people passing in conflicting params

        extra_node_styles = {
            node: get_node_styles(instruction) for node, instruction in instructions
        }

        return self.graphviz(
            extra_node_styles=extra_node_styles, filter=filter, **kwargs
        )

    def graphviz(
        self,
        *,
        hide_parameters=False,
        expand_links=False,
        flatten=False,
        highlight=None,
        filter=None,
        extra_node_styles=None,
    ):
        """
        Generates a graphviz.DiGraph that is suitable for display.

        This requires graphviz to be installed.

        The output can be directly viewed in a Jupyter notebook.
        """
        extra_node_styles = extra_node_styles or {}
        highlight = highlight or []
        dag = self.dag()

        if filter is None:
            filter = dag.nodes()

        unbound = set(self._unbound().keys())

        # Recursively build from the tree
        def create_subgraph(tree, name=None):

            label = name
            if flatten:
                name = f"flat_{name}" if name else None
            else:
                name = f"cluster_{name}" if name else None

            g = graphviz.Digraph(name=name)
            if label:
                g.attr("graph", label=label, fontname="arial", title="")

            for k, v in tree.items():
                if isinstance(v, str):
                    name = v
                    fn = self.raw_function(name)

                    if hide_parameters and name in self._parameters:
                        continue

                    node_styles = dict(
                        style="rounded, filled", fontname="arial", shape="rect"
                    )
                    is_link = fn and getattr(fn, "_is_fn_graph_link", False)

                    if name in highlight:
                        color = "#7dc242"
                    elif name in unbound:
                        color = "red"
                    elif name in self._parameters:
                        color = "lightblue"

                    else:
                        color = "lightgrey"

                    node_styles.update(dict(fillcolor=color))

                    if is_link and expand_links:
                        node_styles.update(dict(fontcolor="darkgrey"))
                    elif is_link:
                        node_styles.update(
                            dict(shape="circle", height="0.2", width="0.2")
                        )

                    node_styles.update(extra_node_styles.get(v, {}))

                    if is_link and not expand_links:
                        label = ""
                    elif flatten:
                        label = name.replace("_", "\n")
                    else:
                        label = k.replace("_", "\n")

                    if name in filter:
                        g.node(name, label=label, **node_styles)
                else:
                    g.subgraph(create_subgraph(v, k))
            return g

        result = create_subgraph(self._build_name_tree())
        result.attr("graph", rankdir="BT")
        for node in self.dag().nodes():
            for _, pred in self._resolve_predecessors(node):
                if (not hide_parameters or pred not in self._parameters) and (
                    pred in filter and node in filter
                ):
                    result.edge(pred, node)

        return result

    def set_source_map(self, source_map):
        """
        Source maps allow you to override the code returned by get source. 

        This is rarely used, and only in esoteric circumstances.
        """
        return self._copy(_source_map=source_map)

    def get_source(self, key):
        """
        Returns the source code that defines this function.
        """
        fn = self.raw_function(key)
        is_parameter = key in self._parameters

        if key in self._source_map:
            return self._source_map[key]
        elif is_parameter:
            return f"{key} = lambda: {self._parameters[key]}"
        elif getattr(fn, "_is_fn_graph_link", False):
            parameter = list(inspect.signature(fn).parameters.keys())[0]
            return f"{key} = lambda {parameter}: {parameter}"
        else:
            try:
                return inspect.getsource(fn)
            except OSError:
                return "Source could not be located"

    def _build_name_tree(self):
        # Build up a tree of the subgraphs
        def recursive_tree():
            return defaultdict(recursive_tree)

        tree = recursive_tree()
        unbound = set(self._unbound().keys())

        for node in [*self.dag().nodes(), *unbound]:
            root = tree
            parts = node.split("__")
            for part in parts[:-1]:
                root = root[part]
            root[parts[-1]] = node
        return tree

    def _unbound(self):

        return {
            k: [t for _, t in v]
            for k, v in groupby(
                sorted(
                    [
                        (arg, key)
                        for key in self._functions
                        for _, arg in self._resolve_predecessors(key)
                        if arg not in self._functions
                    ]
                ),
                key=lambda t: t[0],
            )
        }

    def _resolve_predecessor(self, fname, pname):
        fparts = fname.split("__")[:]
        possible_preds = [
            "__".join(fparts[:i] + [pname]) for i in range(0, len(fparts))
        ]
        possible_preds.reverse()
        for possible in possible_preds:
            if possible in self._functions:
                return possible
        else:
            return possible_preds[0]

    def _resolve_var_predecessors(self, fname, pname):
        fparts = fname.split("__")[:]
        possible_preds = [
            "__".join(fparts[:i] + [pname]) for i in range(0, len(fparts))
        ]
        possible_preds.reverse()

        for possible_prefix in possible_preds:
            for function in self._functions.keys():
                if function.startswith(possible_prefix):
                    key = pname + function[len(possible_prefix) :]

                    yield key, function

    def _resolve_predecessors(self, fname):

        if fname not in self._functions:
            return []

        fn = self._functions[fname]
        sig = signature(fn)
        for key, parameter in sig.parameters.items():
            if parameter.kind in (
                parameter.KEYWORD_ONLY,
                parameter.POSITIONAL_OR_KEYWORD,
                parameter.POSITIONAL_ONLY,
            ):
                resolved_name = self._resolve_predecessor(fname, key)

                if parameter.default is not Parameter.empty:
                    if resolved_name in self._functions:
                        yield key, resolved_name
                else:
                    yield key, resolved_name
            else:
                yield from self._resolve_var_predecessors(fname, key)




import sys
import traceback
from collections import Counter
from enum import Enum
from functools import reduce
from inspect import Parameter, signature
from logging import getLogger

import networkx as nx
from littleutils import ensure_list_if_string

log = getLogger(__name__)


class NodeInstruction(Enum):
    CALCULATE = 1
    RETRIEVE = 2
    IGNORE = 3


def get_execution_instructions(composer, dag, outputs):

    direct_invalid_nodes = {
        node for node in dag if not composer._cache.valid(composer, node)
    }

    invalid_nodes = (
        reduce(
            lambda a, b: a | b,
            [nx.descendants(dag, node) for node in direct_invalid_nodes] + [set()],
        )
        | direct_invalid_nodes
    )

    must_be_retrieved = (
        {
            node
            for node in dag.nodes()
            if any(succ in invalid_nodes for succ in dag.successors(node))
        }
        | set(outputs)
    ) - invalid_nodes

    log.debug("Invalid nodes %s", invalid_nodes)
    log.debug("Retrived Nodes %s", must_be_retrieved)

    execution_order = list(nx.topological_sort(dag))
    execution_instructions = []
    for node in execution_order:
        log.debug(node)
        if node in invalid_nodes:
            node_instruction = NodeInstruction.CALCULATE
        elif node in must_be_retrieved:
            node_instruction = NodeInstruction.RETRIEVE
        else:
            node_instruction = NodeInstruction.IGNORE
        execution_instructions.append((node, node_instruction))

    return execution_instructions


def maintain_cache_consistency(composer):
    dag = composer.dag()

    direct_invalid_nodes = {
        node for node in dag if not composer._cache.valid(composer, node)
    }
    log.debug("Direct invalid nodes %s", direct_invalid_nodes)

    # If a node is invalidate all it's descendents must be made invalid
    indirect_invalid_nodes = (
        reduce(
            lambda a, b: a | b,
            [nx.descendants(dag, node) for node in direct_invalid_nodes] + [set()],
        )
    ) - direct_invalid_nodes
    log.debug("Indirect invalid nodes %s", indirect_invalid_nodes)

    for node in indirect_invalid_nodes:
        composer._cache.invalidate(composer, node)


def coalesce_argument_names(function, predecessor_results):
    sig = signature(function)

    positional_names = []
    args_name = None
    keyword_names = []
    kwargs_name = None

    for key, parameter in sig.parameters.items():
        if parameter.kind in (
            parameter.KEYWORD_ONLY,
            parameter.POSITIONAL_OR_KEYWORD,
            parameter.POSITIONAL_ONLY,
        ):
            if (
                parameter.default is not Parameter.empty
                and key not in predecessor_results
            ):
                continue

            if args_name is None:
                positional_names.append(key)
            else:
                keyword_names.append(key)
        elif parameter.kind == parameter.VAR_POSITIONAL:
            args_name = key
        elif parameter.kind == parameter.VAR_KEYWORD:
            kwargs_name = key

    return positional_names, args_name, keyword_names, kwargs_name


def coalesce_arguments(function, predecessor_results):
    positional_names, args_name, keyword_names, kwargs_name = coalesce_argument_names(
        function, predecessor_results
    )

    positional = [predecessor_results.pop(name) for name in positional_names]
    args = [
        predecessor_results.pop(name)
        for name in list(predecessor_results.keys())
        if name.startswith(args_name)
    ]
    keywords = {name: predecessor_results.pop(name) for name in keyword_names}
    kwargs = {
        name: predecessor_results.pop(name)
        for name in list(predecessor_results)
        if name.startswith(kwargs_name)
    }

    assert len(predecessor_results) == 0

    return positional, args, keywords, kwargs


def calculate_collect_exceptions(
    composer,
    outputs,
    perform_checks=True,
    intermediates=False,
    progress_callback=None,
    raise_immediately=False,
):
    """
    Executes the required parts of the function graph to product results
    for the given outputs.

    Args:
        composer: The composer to calculate
        outputs: list of the names of the functions to calculate
        perform_checks: if true error checks are performed before calculation
        intermediates: if true the results of all functions calculated will be returned
        progress_callback: a callback that is called as the calculation progresses,\
                this be of the form `callback(event_type, details)`

    Returns:
        Tuple: Tuple of (results, exception_info), results is a dictionary of results keyed by 
               function name, exception_info (etype, evalue, etraceback, node) is the information 
               about the exception if there was any.
    """
    outputs = ensure_list_if_string(outputs)

    progress_callback = progress_callback or (lambda *args, **kwargs: None)

    progress_callback("start_calculation", dict(outputs=outputs))

    if perform_checks:
        try:
            for name in outputs:
                if name not in composer._functions:
                    raise Exception(
                        f"'{name}' is not a composed function in this {composer.__class__.__name__} object."
                    )

            for error in composer.check(outputs):
                raise Exception(error)
        except:
            if raise_immediately:
                raise
            else:
                etype, evalue, etraceback = sys.exc_info()

                return {}, (etype, evalue, etraceback, None)

    maintain_cache_consistency(composer)

    # Limit to only the functions we care about
    dag = composer.ancestor_dag(outputs)

    if intermediates:
        outputs = dag.nodes()

    execution_instructions = get_execution_instructions(composer, dag, outputs)
    log.debug(execution_instructions)

    # Results store
    results = {}

    # Number of time a functions results still needs to be accessed
    remaining_usage_counts = Counter(pred for pred, _ in dag.edges())
    progress_callback(
        "prepared_calculation",
        dict(execution_instructions=execution_instructions, execution_graph=dag),
    )
    log.debug("Starting execution")
    for node, instruction in execution_instructions:
        progress_callback(
            "start_step", dict(name=node, execution_instruction=instruction)
        )
        try:
            predecessors = list(composer._resolve_predecessors(node))

            if instruction == NodeInstruction.IGNORE:
                log.debug("Ignoring function '%s'", node)

            elif instruction == NodeInstruction.RETRIEVE:
                log.debug("Retrieving function '%s'", node)
                try:
                    progress_callback("start_cache_retrieval", dict(name=node))
                    results[node] = composer._cache.get(composer, node)
                finally:
                    progress_callback("end_cache_retrieval", dict(name=node))
            else:
                log.debug("Calculating function '%s'", node)
                function = composer._functions[node]

                # Pull up arguments
                predecessor_results = {
                    parameter: results[pred] for parameter, pred in predecessors
                }
                positional, args, keywords, kwargs = coalesce_arguments(
                    function, predecessor_results
                )

                try:
                    progress_callback("start_function", dict(name=node))
                    result = function(*positional, *args, **keywords, **kwargs)
                except Exception as e:
                    if raise_immediately:
                        raise
                    else:
                        etype, evalue, etraceback = sys.exc_info()

                        return results, (etype, evalue, etraceback, node)
                finally:
                    progress_callback("end_function", dict(name=node))

                results[node] = result

                try:
                    progress_callback("start_cache_store", dict(name=node))
                    composer._cache.set(composer, node, result)
                finally:
                    progress_callback("end_cache_store", dict(name=node))

            # Eject results from memory once the are not needed
            remaining_usage_counts.subtract([pred for _, pred in predecessors])
            ready_to_eject = [
                key
                for key, value in remaining_usage_counts.items()
                if value == 0 and key not in outputs
            ]
            for key in ready_to_eject:
                assert key not in outputs
                remaining_usage_counts.pop(key)
                results.pop(key, "not_found")
        finally:
            progress_callback(
                "end_step",
                dict(
                    name=node,
                    execution_instruction=instruction,
                    result=results.get(node),
                ),
            )

    # We should just be left with the results
    return results, None


def calculate(*args, **kwargs):
    """
    Executes the required parts of the function graph to product results
    for the given outputs.

    Args:
        composer: The composer to calculate
        outputs: list of the names of the functions to calculate
        perform_checks: if true error checks are performed before calculation
        intermediates: if true the results of all functions calculated will be returned
        progress_callback: a callback that is called as the calculation progresses,\
                this be of the form `callback(event_type, details)`

    Returns:
        Dictionary: Dictionary of results keyed by function name
    """
    results, _ = calculate_collect_exceptions(*args, raise_immediately=True, **kwargs)
    return results



"Examples of various different supported plotting libraries."
#%%
import math

import matplotlib.pylab as plt
import plotly.express as px
import seaborn as sns
from fn_graph import Composer


def matplotlib_plot(data):
    fig2, ax = plt.subplots()
    ax.plot(data.total_bill, data.tip, "o")
    return fig2


def seaborn_plot(data):

    return sns.relplot(
        x="total_bill",
        y="tip",
        col="time",
        hue="smoker",
        style="smoker",
        size="size",
        data=data,
    )


def pandas_plot(data):
    return data.total_bill.plot.hist()


def plotly_plot(data):
    return px.scatter(
        data,
        x="total_bill",
        y="tip",
        color="sex",
        facet_row="time",
        facet_col="day",
        trendline="ols",
        size="size",
        symbol="smoker",
    )


f = (
    Composer()
    .update_parameters(data=sns.load_dataset("tips"))
    .update(matplotlib_plot, seaborn_plot, pandas_plot, plotly_plot)
)



"""
This example generates some very large objects mainly to test how well they ware handled by visual frontends
"""
# %%
from fn_graph import Composer
from pathlib import Path
import pandas as pd
import numpy as np
import plotly.express as px


def dataframe(dataframe_width, dataframe_length):
    return pd.DataFrame(
        {
            f"dimension_{i}": np.random.random(dataframe_length)
            for i in range(dataframe_width)
        }
    )


def dictionary(dataframe: pd.DataFrame):
    return dataframe.to_dict()


def to_set(dictionary: dict):
    return {f"{k}-{v}" for k, dimension in dictionary.items() for v in dimension}


f = (
    Composer()
    .update_parameters(dataframe_length=100, dataframe_width=10)
    .update(dataframe, dictionary, set=to_set)
)



from random import randint, choice
from fn_graph import Composer
from textwrap import dedent


def check_all_equal(values):
    return any([a == b for a, b in list(zip(values[:-1], values[1:]))])


def check_results_equal(results):
    if not check_all_equal([set(result.keys()) for result in results]):
        raise Exception(f"Keys differ")

    for key in results[0]:
        result_values = [result[key] for result in results]
        if not check_all_equal(result_values):
            raise Exception(f"Difference found in {key}, results: {result_values}")


def compare_composer_results(root, composers):
    nodes = root.dag().nodes()
    for j in range(10):
        outputs = set((choice(list(nodes)) for k in range(randint(1, len(nodes)))))
        intermediates = randint(0, 1) == 1
        results = [
            composer.calculate(outputs, intermediates=intermediates)
            for composer in composers
        ]
        check_results_equal(results)

        for composer in composers:
            composer.cache_invalidate(choice(list(nodes)))


def generate_random_graph(graph_size=42):

    functions = []

    def function_0():
        return 42

    functions.append(function_0)

    for i in range(1, graph_size):
        fn_name = f"function_{i}"
        num_args = randint(0, min(5, i - 1))
        arg_names = set()

        while len(arg_names) < num_args:
            arg_name = f"function_{randint(0, i-1)}"
            if arg_name not in arg_names:
                arg_names.add(arg_name)

        if arg_names:
            body = " + ".join(arg_names)
        else:
            body = str(randint(0, 100))

        exec(
            dedent(
                f"""
            def {fn_name}({', '.join(sorted(arg_names))}):
                return {body}
            """
            )
        )
        functions.append(locals()[fn_name])

    return Composer().update(*functions)



from fn_graph import Composer


def function_0():
    return 42


def function_1():
    return 43


def function_2(function_1):
    return function_1


def function_3(function_1):
    return function_1


def function_4(function_0):
    return function_0


def function_5(function_2):
    return function_2


def function_6(function_0, function_2, function_3, function_5):
    return function_3 + function_2 + function_5 + function_0


def function_7(function_2, function_3, function_5, function_6):
    return function_3 + function_2 + function_6 + function_5


# making these links

# def function_8(function_7):
#     return function_7


# def function_9(function_5):
#     return function_5


def function_10(function_1, function_3, function_6, function_7, function_9):
    return function_3 + function_1 + function_6 + function_7 + function_9


def function_11(function_1, function_4):
    return function_4 + function_1


def function_12(function_3, function_5, function_6, function_8, function_9):
    return function_3 + function_6 + function_8 + function_9 + function_5


def function_13(function_0, function_12, function_4, function_8):
    return function_12 + function_4 + function_8 + function_0


def function_14(function_7, function_9):
    return function_7 + function_9


def function_15(function_0, function_14, function_3, function_7):
    return function_3 + function_7 + function_14 + function_0


def function_16(function_0, function_1, function_11, function_3, function_6):
    return function_11 + function_3 + function_1 + function_6 + function_0


def function_17(function_2, function_8):
    return function_2 + function_8


def function_18(function_10, function_13, function_14, function_4):
    return function_4 + function_10 + function_14 + function_13


def function_19(function_10, function_7):
    return function_10 + function_7


def function_20(function_15):
    return function_15


def function_21(function_4):
    return function_4


def function_22(function_10, function_15, function_9):
    return function_15 + function_10 + function_9


def function_23(function_15, function_6):
    return function_15 + function_6


def function_24(function_1, function_12, function_18, function_23, function_5):
    return function_12 + function_1 + function_18 + function_23 + function_5


def function_25(function_14, function_18):
    return function_18 + function_14


def function_26(function_0, function_14, function_2):
    return function_2 + function_14 + function_0


def function_27(function_26):
    return function_26


def function_28(function_17, function_2, function_24, function_25, function_9):
    return function_17 + function_25 + function_24 + function_2 + function_9


def function_29():
    return 29


def function_30(function_13, function_25):
    return function_25 + function_13


def function_31(function_1, function_14, function_15, function_29, function_7):
    return function_15 + function_1 + function_29 + function_7 + function_14


def function_32(function_12, function_22, function_25):
    return function_12 + function_22 + function_25


def function_33(function_3):
    return function_3


def function_34(function_25, function_33, function_4, function_5, function_6):
    return function_4 + function_25 + function_6 + function_33 + function_5


def function_35(function_14, function_17, function_21):
    return function_17 + function_21 + function_14


def function_36(function_33, function_6):
    return function_33 + function_6


def function_37():
    return 19


def function_38(function_33, function_34):
    return function_34 + function_33


def function_39(function_1, function_10, function_25):
    return function_25 + function_1 + function_10


def function_40(function_21, function_24, function_33):
    return function_21 + function_33 + function_24


def function_41(function_18):
    return function_18


#%%

scope = locals()
functions = [scope[f"function_{i}"] for i in range(42) if f"function_{i}" in scope]
large_graph = (
    Composer().update(*functions).link(function_8="function_7", function_9="function_5")
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fn_graph', 'fn_graph.examples', 'fn_graph.tests']

package_data = \
{'': ['*']}

install_requires = \
['graphviz>=0.13.2,<0.14.0',
 'joblib>=0.14.1,<0.15.0',
 'littleutils>=0.2.1,<0.3.0',
 'networkx>=2.4,<3.0']

setup_kwargs = {
    'name': 'fn-graph',
    'version': '0.8.1',
    'description': 'Manage, maintain and reuse complex function graphs without the hassle.',
    'long_description': '# Fn Graph\n\nLight weight function pipelines for python.\n\n## Overview\n\n`fn_graph` is trying to solve a number of problems in the python data-science/modelling domain, as well as making it easier to put such models into production.\n\nIt aims to:\n\n1. Make moving between the analyst space to production. amd back, simpler and less error prone.\n2. Make it easy to view the intermediate results of computations to easily diagnose errors.\n3. Solve common analyst issues like creating reusable, composable pipelines and caching results.\n4. Visualizing models in an intuitive way.\n\nThere is an associated visual studio you should check out at https://github.com/BusinessOptics/fn_graph_studio/.\n\n## Documentation\n\nPlease find detailed documentation at https://fn-graph.readthedocs.io/\n\n## Installation\n\n```\npip install fn_graph\n```\n\nYou will need to have graphviz and the development packages installed. On ubuntu you can install these with:\n\n```\nsudo apt-get install graphviz graphviz-dev\n```\n\nOtherwise see the [pygraphviz documentation](http://pygraphviz.github.io/documentation/pygraphviz-1.5/install.html).\n\n## Outline of the problem\n\nIn complex domains, like financial modelling or data-science, modellers and programmers often have complex logic that needs to be re-used in multiple environments. For example a financial analyst may write a model in a notebook, which then must be placed in a production environment (normally through rewriting the notebook which was not written in a modular or easy to use manner), after which changes need to made, which must be done in the notebook environment, which then need to be moved back into production. This is the stuff of [nightmare fuel](https://www.urbandictionary.com/define.php?term=nightmare%20fuel) for everyone involved.\n\nTo make this process easier the best option is to break functionality into small reusable functions, that can be referenced in both the notebook environment and the production environment. Now you unfortunately you have this bag of small functions whose relationships you have to manage. There are few ways to do this. Assume we have the functions below:\n\n```python\ndef get_a():\n    return 5\n\ndef get_b(a):\n    return a * 5\n\ndef get_c(a, b):\n    return a * b\n```\n\n**Option 1 - Directly compose at call site:**\n\n```python\na = get_a()\nb = get_b(a)\nc = get_c(a, b)\n```\n\n_Pros:_ The modeller can easily see intermediate results\\\n_Cons:_ Now this potentially complex network of function calls has to be copied and pasted between notebooks and production\n\n**Option 2 - Wrap it up in a function that pass around**\n\n```python\ndef composed_c():\n    a = get_a()\n    b = get_b(a)\n    return  get_c(a, b)\n```\n\n_Pros:_ This is easy to reference from both the notebook and production\\\n_Cons:_ The modeller cannot see the intermediate results, so it can be difficult to debug (this is a big problem if you are working with big multi-dimensional objects that you may want to visualize, a debugger does not cut it)\n\n**Option 3 - Directly call functions from each other**\n\n```python\ndef get_a():\n    return 5\n\ndef get_b(a):\n    return get_a() * 5\n\ndef get_c():\n    return get_a() * get_b()\n```\n\n_Pros:_ This is easy to reference from both the notebook and production\\\n_Cons:_ The modeller cannot see the intermediate results and functions cannot be reused. Functions are called multiple times.\n\nNone of these are great. Fn Graph would solve it like this.\n\n```python\nfrom fn_graph import Composer\n\ndef a():\n    return 5\n\ndef b(a):\n    return a * 5\n\ndef c(a, b):\n    return a * b\n\ncomposer = Composer().update(a, b, c)\n\n# Call any result\ncomposer.c() # 125\ncomposer.a() # 5\n\ncomposer.graphviz()\n```\n\n![Graph of composer](intro.gv.png)\n\nThe composer can then be easily passed around in both the production and notebook environment. It can do much more than this.\n\n## Features\n\n- Manage complex function graphs, including using namespaces.\n- Update composers to gradually build more and more complex logic.\n- Enable incredible function reuse.\n- Visualize logic to make knowledge sharing easier.\n- Perform graph operations on composers to dynamically rewire your logic.\n- Manage calculation life cycle, with hooks, and have access to all intermediary calculations.\n- Cache results, either within a single session, or between sessions in development mode. Using the development cache intelligently invalidate the cache when code changes .\n\n## Similar projects\n\n**Dask**\n\nDask is a light-weight parallel computing library. Importantly it has a Pandas compliant interface. You may want to use Dask inside FnGraph.\n\n**Airflow**\n\nAirflow is a task manager. It is used to run a series of generally large tasks in an order that meets their dependencies, potentially over multiple machines. It has a whole scheduling and management apparatus around it. Fn Graph is not trying to do this. Fn Graph is about making complex logic more manageable, and easier to move between development and production. You may well want to use Fn Graph inside your airflow tasks.\n\n**Luigi**\n\n> Luigi is a Python module that helps you build complex pipelines of batch jobs. It handles dependency resolution, workflow management, visualization etc. It also comes with Hadoop support built in.\n\nLuigi is about big batch jobs, and managing the distribution and scheduling of them. In the same way that airflow works ate a higher level to FnGraph, so does luigi.\n\n**d6tflow**\n\nd6tflow is very similar to FnGraph. It is based on Luigi. The primary difference is the way the function graphs are composed. d6tflow graphs can be very difficult to reuse (but do have some greater flexibility). It also allows for parallel execution. FnGraph is trying to make very complex pipelines or very complex models easier to mange, build, and productionise.\n',
    'author': 'James Saunders',
    'author_email': 'james@businessoptics.biz',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/BusinessOptics/fn_graph',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import hashlib
import inspect
import json
import pickle
from io import BytesIO
from logging import getLogger
from pathlib import Path


log = getLogger(__name__)


def fn_value(fn):
    if getattr(fn, "_is_fn_graph_link", False):
        return ",".join(inspect.signature(fn).parameters.keys())
    else:
        return inspect.getsource(fn)


def hash_fn(composer, key, use_id=False):
    value = (
        composer._parameters[key][1]
        if key in composer._parameters
        else composer._functions[key]
    )

    log.debug(f"Fn Value %s: %s", key, value)

    if use_id:
        return id(value)

    if callable(value):
        buffer = fn_value(value).encode("utf-8")
        return hashlib.sha256(buffer).digest()
    else:
        buffer = BytesIO()
        pickle.dump(value, buffer)
        return hashlib.sha256(buffer.getvalue()).digest()


class NullCache:
    """
    Performs no caching.

    Used as a base class for other caches to indicate that they all have the same signature.
    """

    def valid(self, composer, key):
        return False

    def get(self, composer, key):
        pass

    def set(self, composer, key, value):
        pass

    def invalidate(self, composer, key):
        pass


class SimpleCache(NullCache):
    """
    Stores results in memory, performs no automatic invalidation.

    Set hash_parameters=False to only check the object identities of parameters
    when checking cache validity. This is slightly less robust to buggy usage
    but is much faster for big parameter objects.

    DO NOT USE THIS IN DEVELOPMENT OR A NOTEBOOK!
    """

    def __init__(self, hash_parameters=False):
        self.cache = {}
        self.hashes = {}
        self.hash_parameters = hash_parameters

    def _hash(self, composer, key):
        return hash_fn(composer, key, use_id=not self.hash_parameters)

    def valid(self, composer, key):
        log.debug("Length %s", len(composer._functions))
        if key in self.hashes:
            current_hash = self._hash(composer, key)
            stored_hash = self.hashes[key]
            matches = current_hash == stored_hash
            log.debug(f"hash test {key} {matches}: {current_hash} {stored_hash}")
            return matches
        else:
            log.debug(f"cache test {key} {key in self.cache}")
            return key in self.cache

    def get(self, composer, key):
        return self.cache[key]

    def set(self, composer, key, value):
        self.cache[key] = value
        if key in composer.parameters():
            self.hashes[key] = self._hash(composer, key)

    def invalidate(self, composer, key):

        if key in self.cache:
            del self.cache[key]

        if key in self.hashes:
            del self.hashes[key]


class DevelopmentCache(NullCache):
    """
    Store cache on disk, analyses the coe for changes and performs automatic
    invalidation.

    This is only for use during development! DO NOT USE THIS IN PRODUCTION!

    The analysis of code changes is limited, it assumes that all functions are
    pure, and tht there have been no important changes in the outside environment,
    like a file that has been changed,
    """

    def __init__(self, name, cache_dir):
        self.name = name

        if cache_dir is None:
            cache_dir = ".fn_graph_cache"

        self.cache_dir = Path(cache_dir)
        self.cache_root = self.cache_dir / name
        self.cache_root.mkdir(parents=True, exist_ok=True)

    def valid(self, composer, key):
        pickle_file_path = self.cache_root / f"{key}.data"
        info_file_path = self.cache_root / f"{key}.info.json"
        fn_hash_path = self.cache_root / f"{key}.fn.hash"

        exists = pickle_file_path.exists() and info_file_path.exists()

        log.debug(
            "Checking development cache '%s' for key '%s': exists = %s",
            self.name,
            key,
            exists,
        )

        if not exists:
            return False

        current_hash = hash_fn(composer, key)
        with open(fn_hash_path, "rb") as f:
            previous_hash = f.read()

        if current_hash != previous_hash:
            log.debug(
                "Function change detected in development cache '%s' for key '%s'",
                self.name,
                key,
            )
            return False

        return True

    def get(self, composer, key):
        log.debug("Retrieving from development cache '%s' for key '%s'", self.name, key)

        pickle_file_path = self.cache_root / f"{key}.data"
        info_file_path = self.cache_root / f"{key}.info.json"

        with open(info_file_path, "r") as f:
            information = json.load(f)

        format = information["format"]

        with open(pickle_file_path, "rb") as f:
            if format == "pickle":
                return pickle.load(f)
            elif format == "pandas-parquet":
                import pandas as pd

                return pd.read_parquet(f)
            else:
                raise Exception(f"Unknown caching fn_graph format: {format}")

    def set(self, composer, key, value):

        log.debug("Writing to development cache '%s' for key '%s'", self.name, key)
        pickle_file_path = self.cache_root / f"{key}.data"
        info_file_path = self.cache_root / f"{key}.info.json"
        fn_hash_path = self.cache_root / f"{key}.fn.hash"

        # This is a low-fi way to checK the type without adding requirements
        # I am concerned it is fragile
        if str(type(value)) == "<class 'pandas.core.frame.DataFrame'>":
            format = "pandas-parquet"
        else:
            format = "pickle"

        saved = False
        if format == "pandas-parquet":
            try:
                with open(pickle_file_path, "wb") as f:
                    value.to_parquet(f)
                saved = True
            except:
                saved = False

        if not saved:
            format = "pickle"
            with open(pickle_file_path, "wb") as f:
                pickle.dump(value, f)

        with open(fn_hash_path, "wb") as f:
            f.write(hash_fn(composer, key))

        with open(info_file_path, "w") as f:
            json.dump({"format": format}, f)

    def invalidate(self, composer, key):
        log.debug("Invalidation in development cache '%s' for key '%s'", self.name, key)
        paths = [
            self.cache_root / f"{key}.data",
            self.cache_root / f"{key}.fn.hash",
            self.cache_root / f"{key}.info.json",
        ]

        for path in paths:
            if path.exists():
                path.unlink()



"""
A simple example that works out the market capitalization fo a couple of stocks.
"""
# %%
from fn_graph import Composer
from pathlib import Path
import pandas as pd
import numpy as np
import plotly.express as px

data_path = Path(__file__).parent


def share_prices():
    """
    Load the share price data
    """
    return pd.read_csv(data_path / "share_prices.csv", parse_dates=["datetime"])


def shares_in_issue():
    """
    Load the shares issued data
    """
    return pd.read_csv(data_path / "shares_in_issue.csv", parse_dates=["datetime"])


def daily_share_prices(share_prices):
    """
    Ensure that every day has the full set of share prices
    """
    return (
        share_prices.groupby("share_code")
        .apply(lambda df: df.set_index("datetime").resample("1D").ffill().reset_index())
        .reset_index(drop=True)
        .sort_values(by=["datetime", "share_code"])
    )


def market_cap(daily_share_prices, shares_in_issue):
    """
    Merge the datasets intelligently over time and calculate market cap
    """
    return pd.merge_asof(
        daily_share_prices, shares_in_issue, on="datetime", by="share_code"
    ).assign(market_cap=lambda df: df.share_price * df.shares_in_issue)


def total_market_cap(market_cap):
    """
    Workout the total market cap
    """
    return market_cap.groupby("datetime", as_index=False).market_cap.sum()


def total_market_cap_change(total_market_cap, swing_threshold):
    """
    Calculate the changes in market cap
    """
    return total_market_cap.assign(
        market_cap_change=lambda df: df.market_cap.diff()
    ).assign(
        change_classification=lambda df: np.where(
            np.abs(df.market_cap_change) > swing_threshold, "large", "small"
        )
    )


def plot_market_caps(market_cap):
    """
    Plot the individual market caps
    """
    return px.area(
        market_cap,
        x="datetime",
        y="market_cap",
        facet_row="share_code",
        color="share_code",
    )


def plot_total_market_cap(total_market_cap):
    """
    Plot the total market cap
    """
    return px.line(total_market_cap, x="datetime", y="market_cap")


def plot_market_cap_changes(total_market_cap_change):
    """
    Plot the market cap changes
    """
    return px.bar(
        total_market_cap_change,
        x="datetime",
        y="market_cap_change",
        color="change_classification",
    )


f = (
    Composer()
    .update_parameters(swing_threshold=0.1 * 10 ** 12)
    .update(
        share_prices,
        daily_share_prices,
        shares_in_issue,
        market_cap,
        total_market_cap,
        total_market_cap_change,
        plot_market_caps,
        plot_total_market_cap,
        plot_market_cap_changes,
    )
)
